from .base import Loss
from .base import functional as F

SMOOTH = 1e-5


class JaccardLoss(Loss):
    r"""Creates a criterion to measure Jaccard loss:

    .. math:: L(A, B) = 1 - \frac{A \cap B}{A \cup B}

    Args:
        class_weights: Array (``np.array``) of class weights (``len(weights) = num_classes``).
        class_indexes: Optional integer or list of integers, classes to consider, if ``None`` all classes are used.
        per_image: If ``True`` loss is calculated for each image in batch and then averaged,
            else loss is calculated for the whole batch.
        smooth: Value to avoid division by zero.

    Returns:
         A callable ``jaccard_loss`` instance. Can be used in ``model.compile(...)`` function
         or combined with other losses.

    Example:

    .. code:: python

        loss = JaccardLoss()
        model.compile('SGD', loss=loss)
    """

    def __init__(self, class_weights=None, class_indexes=None, per_image=False, smooth=SMOOTH):
        super().__init__(name='jaccard_loss')
        self.class_weights = class_weights if class_weights is not None else 1
        self.class_indexes = class_indexes
        self.per_image = per_image
        self.smooth = smooth

    def __call__(self, gt, pr):
        return 1 - F.iou_score(
            gt,
            pr,
            class_weights=self.class_weights,
            class_indexes=self.class_indexes,
            smooth=self.smooth,
            per_image=self.per_image,
            threshold=None,
            **self.submodules
        )


class DiceLoss(Loss):
    r"""Creates a criterion to measure Dice loss:

    .. math:: L(precision, recall) = 1 - (1 + \beta^2) \frac{precision \cdot recall}
        {\beta^2 \cdot precision + recall}

    The formula in terms of *Type I* and *Type II* errors:

    .. math:: L(tp, fp, fn) = \frac{(1 + \beta^2) \cdot tp} {(1 + \beta^2) \cdot fp + \beta^2 \cdot fn + fp}

    where:
         - tp - true positives;
         - fp - false positives;
         - fn - false negatives;

    Args:
        beta: Float or integer coefficient for precision and recall balance.
        class_weights: Array (``np.array``) of class weights (``len(weights) = num_classes``).
        class_indexes: Optional integer or list of integers, classes to consider, if ``None`` all classes are used.
        per_image: If ``True`` loss is calculated for each image in batch and then averaged,
        else loss is calculated for the whole batch.
        smooth: Value to avoid division by zero.

    Returns:
        A callable ``dice_loss`` instance. Can be used in ``model.compile(...)`` function`
        or combined with other losses.

    Example:

    .. code:: python

        loss = DiceLoss()
        model.compile('SGD', loss=loss)
    """

    def __init__(self, beta=1, class_weights=None, class_indexes=None, per_image=False, smooth=SMOOTH):
        super().__init__(name='dice_loss')
        self.beta = beta
        self.class_weights = class_weights if class_weights is not None else 1
        self.class_indexes = class_indexes
        self.per_image = per_image
        self.smooth = smooth

    def __call__(self, gt, pr):
        return 1 - F.f_score(
            gt,
            pr,
            beta=self.beta,
            class_weights=self.class_weights,
            class_indexes=self.class_indexes,
            smooth=self.smooth,
            per_image=self.per_image,
            threshold=None,
            **self.submodules
        )


class BinaryCELoss(Loss):
    """Creates a criterion that measures the Binary Cross Entropy between the
    ground truth (gt) and the prediction (pr).

    .. math:: L(gt, pr) = - gt \cdot \log(pr) - (1 - gt) \cdot \log(1 - pr)

    Returns:
        A callable ``binary_crossentropy`` instance. Can be used in ``model.compile(...)`` function
        or combined with other losses.

    Example:

    .. code:: python

        loss = BinaryCELoss()
        model.compile('SGD', loss=loss)
    """

    def __init__(self):
        super().__init__(name='binary_crossentropy')

    def __call__(self, gt, pr):
        return F.binary_crossentropy(gt, pr, **self.submodules)


class CategoricalCELoss(Loss):
    """Creates a criterion that measures the Categorical Cross Entropy between the
    ground truth (gt) and the prediction (pr).

    .. math:: L(gt, pr) = - gt \cdot \log(pr)

    Args:
        class_weights: Array (``np.array``) of class weights (``len(weights) = num_classes``).
        class_indexes: Optional integer or list of integers, classes to consider, if ``None`` all classes are used.

    Returns:
        A callable ``categorical_crossentropy`` instance. Can be used in ``model.compile(...)`` function
        or combined with other losses.

    Example:

    .. code:: python

        loss = CategoricalCELoss()
        model.compile('SGD', loss=loss)
    """

    def __init__(self, class_weights=None, class_indexes=None):
        super().__init__(name='categorical_crossentropy')
        self.class_weights = class_weights if class_weights is not None else 1
        self.class_indexes = class_indexes

    def __call__(self, gt, pr):
        return F.categorical_crossentropy(
            gt,
            pr,
            class_weights=self.class_weights,
            class_indexes=self.class_indexes,
            **self.submodules
        )


class CategoricalFocalLoss(Loss):
    r"""Creates a criterion that measures the Categorical Focal Loss between the
    ground truth (gt) and the prediction (pr).

    .. math:: L(gt, pr) = - gt \cdot \alpha \cdot (1 - pr)^\gamma \cdot \log(pr)

    Args:
        alpha: Float or integer, the same as weighting factor in balanced cross entropy, default 0.25.
        gamma: Float or integer, focusing parameter for modulating factor (1 - p), default 2.0.
        class_indexes: Optional integer or list of integers, classes to consider, if ``None`` all classes are used.

    Returns:
        A callable ``categorical_focal_loss`` instance. Can be used in ``model.compile(...)`` function
        or combined with other losses.

    Example:

        .. code:: python

            loss = CategoricalFocalLoss()
            model.compile('SGD', loss=loss)
    """

    def __init__(self, alpha=0.25, gamma=2., class_indexes=None):
        super().__init__(name='focal_loss')
        self.alpha = alpha
        self.gamma = gamma
        self.class_indexes = class_indexes

    def __call__(self, gt, pr):
        return F.categorical_focal_loss(
            gt,
            pr,
            alpha=self.alpha,
            gamma=self.gamma,
            class_indexes=self.class_indexes,
            **self.submodules
        )


class BinaryFocalLoss(Loss):
    r"""Creates a criterion that measures the Binary Focal Loss between the
    ground truth (gt) and the prediction (pr).

    .. math:: L(gt, pr) = - gt \alpha (1 - pr)^\gamma \log(pr) - (1 - gt) \alpha pr^\gamma \log(1 - pr)

    Args:
        alpha: Float or integer, the same as weighting factor in balanced cross entropy, default 0.25.
        gamma: Float or integer, focusing parameter for modulating factor (1 - p), default 2.0.

    Returns:
        A callable ``binary_focal_loss`` instance. Can be used in ``model.compile(...)`` function
        or combined with other losses.

    Example:

    .. code:: python

        loss = BinaryFocalLoss()
        model.compile('SGD', loss=loss)
    """

    def __init__(self, alpha=0.25, gamma=2.):
        super().__init__(name='binary_focal_loss')
        self.alpha = alpha
        self.gamma = gamma

    def __call__(self, gt, pr):
        return F.binary_focal_loss(gt, pr, alpha=self.alpha, gamma=self.gamma, **self.submodules)


# aliases
jaccard_loss = JaccardLoss()
dice_loss = DiceLoss()

binary_focal_loss = BinaryFocalLoss()
categorical_focal_loss = CategoricalFocalLoss()

binary_crossentropy = BinaryCELoss()
categorical_crossentropy = CategoricalCELoss()

# loss combinations
bce_dice_loss = binary_crossentropy + dice_loss
bce_jaccard_loss = binary_crossentropy + jaccard_loss

cce_dice_loss = categorical_crossentropy + dice_loss
cce_jaccard_loss = categorical_crossentropy + jaccard_loss

binary_focal_dice_loss = binary_focal_loss + dice_loss
binary_focal_jaccard_loss = binary_focal_loss + jaccard_loss

categorical_focal_dice_loss = categorical_focal_loss + dice_loss
categorical_focal_jaccard_loss = categorical_focal_loss + jaccard_loss



# -*- coding: utf-8 -*-
"""
Created on Wed Jul 22 16:24:31 2020

@author: Koerber
"""

from dl.method.mode import LearningMode, dlMode
from dl.loss.classification_losses import ClassificationLosses
from dl.metric.classification_metrics import ClassificationMetrics
from utils.shapes.ImageLabel import ImageLabel, saveImageLabel, loadImageLabel, drawImageLabel, extractImageLabel
from utils.shapes.Contour import removeBackground
from tensorflow.keras.utils import to_categorical
from dl.models.classification_models.tfkeras import Classifiers
from tensorflow.keras.models import load_model
import tensorflow as tf
import numpy as np
import cv2

class Classification(LearningMode):
    def __init__(self, parent):
        super(Classification,self).__init__(parent)
        self.type = dlMode.Classification
        self.loss = ClassificationLosses(parent)
        self.metric = ClassificationMetrics(parent)
        self.preprocessingfnc = None
        self.backbone = 'resnet50'
        self.architecture = 'standard'
        
    def LoadLabel(self, filename, height, width, rmBackground = False):
        # we need an image here to perform background subtraction
        label = np.zeros((height, width, 1), np.uint8)
        imagelabel, bg = loadImageLabel(filename)
        label = drawImageLabel(label, imagelabel, bg)
        if rmBackground:
            return removeBackground(label)
        return label
        
    def prepreprocessLabel(self, label):
        # convert image to 1D array
        return np.max(label).reshape(1)
    
    def preprocessLabel(self, label):
        if self.parent.NumClasses > 2:
            label = to_categorical(label, num_classes=self.parent.NumClasses)
        else:
            label = label.astype('float')
        return label

    def setBackbone(self, backbone):
        self.backbone = backbone

    def setArchitecture(self, architecture):
        pass

    def getBackbones(self):
        return Classifiers.models_names()

    def getArchitectures(self):
        return ['standard']

    def getModel(self, n_classes, channels):
        backboneModel, self.preprocessingfnc = Classifiers.get(self.backbone)
        addextraInputLayer = False
        if self.pretrained:
            weights = 'imagenet'
            size = (None, None, 3)
            if channels != 3:
                addextraInputLayer = True 
        else:
            weights = None
            self.preprocessingfnc = None  
            size = (None, None,channels)


        if 'nasnet' in self.backbone:
            if self.pretrained:
                size = (224, 224,3)
                # self.parent.augmentation.outputwidth = 224
                # self.parent.augmentation.outputheight = 224
            else:
                size = (self.parent.augmentation.outputwidth, self.parent.augmentation.outputheight,channels)

        basemodel = backboneModel(input_shape=size, weights=weights, include_top=False)

        if addextraInputLayer:
            input = tf.keras.layers.Input(shape=(None, None, channels))
            layer0 = tf.keras.layers.Conv2D(3, (1, 1))(input) 
            out = basemodel(layer0)
            basemodel = tf.keras.models.Model(input, out, name=basemodel.name)

        if n_classes > 2:
            out_activation = 'softmax'
        else:
            out_activation =  'sigmoid'

        # we cut here and pool, to rebuild models like vgg exactly, fc layers (a 4096, 4096 and 1000 units in case of vgg) should be added
        x = tf.keras.layers.GlobalAveragePooling2D()(basemodel.output)
        output = tf.keras.layers.Dense(n_classes, activation=out_activation)(x)
        self.setModelInputSize()
        return tf.keras.models.Model(inputs=[basemodel.input], outputs=[output])
    
    def setModelInputSize(self):
        if 'nasnet' in self.backbone:
            if self.pretrained:
                self.parent.augmentation.outputwidth = 224
                self.parent.augmentation.outputheight = 224

    def loadmodel(self, path):
        if self.pretrained:
            _, self.preprocessingfnc = Classifiers.get(self.backbone)
        else:
            self.preprocessingfnc = None
        self.setModelInputSize()
        return load_model(path)


    def extractShapesFromPrediction(self, prediction, unused1, unused2):
        return extractImageLabel(prediction)
    
    def saveShapes(self, contours, path):
        saveImageLabel(contours, path)
    
    def PredictImage(self, image):
        if not self.parent.initialized or image is None:
            return None

        width, height = self.getImageSize4ModelInput(None,None)
        image = cv2.resize(image, (width, height))
        
        results = []
        origImage = image
        
        pred = self.Predict(image)
        results.append(pred)
        
        if self.parent.tta:
            for i in range(5):               
                if i < 2:
                    flip = i
                    image = np.flip(origImage.copy(),flip)
                    res = self.Predict(image)
                    results.append(res)
                else:
                    axes = (0,1)
                    iaxes = (1,0)
                    image = np.rot90(origImage.copy(), k = i-1, axes=axes)
                    res = self.Predict(image)
                    results.append(res)

        results = np.stack(results, axis = 0)
        result = np.mean(results, axis = 0)

        if self.parent.NumClasses > 2:
            result = np.argmax(result)
        else:
            result[result>0.5] = 1
            result[result<=0.5] = 0  
        return result
    
    def Predict(self,image):
        image = self.parent.data.preprocessImage(image)
        image = image[np.newaxis, ...]
        return self.parent.Model.predict(self.parent.augmentation.checkModelSizeCompatibility(image))
        
    
    def resizeLabel(self, label, shape):
        return cv2.resize(label, shape, interpolation = cv2.INTER_NEAREST)

    def countLabelWeight(self, label):
        return np.zeros((self.parent.NumClasses_real), dtype = np.float)

    def getImageSize4ModelInput(self, inputWidth, inputHeight):
        width = self.parent.augmentation.outputwidth
        height = self.parent.augmentation.outputheight
        return width, height


# -*- coding: utf-8 -*-
"""
Created on Tue Jun  9 11:14:06 2020

@author: Koerber
"""
from math import ceil
from enum import Enum
from tensorflow.keras.callbacks import LearningRateScheduler, ReduceLROnPlateau


class lrMode(Enum):
    Constant = 0
    ReduceOnPlateau = 1
    Schedule = 2
    


class LearningRateSchedule():
    def __init__(self, parent):
        self.parent = parent
        self.ReduceEveryXEpoch = 25
        self.ReductionFactor = 0.5
        self.Mode = lrMode.Constant
        self.minlr = 0.000001
        
    def setMode(self, mode):
        self.Mode = mode
        if mode == lrMode.ReduceOnPlateau:
            self.ReduceEveryXEpoch = 5
            self.ReductionFactor = 0.2

        if mode == lrMode.Schedule:
            self.ReduceEveryXEpoch = 25
            self.ReductionFactor = 0.5

            
        
    def LearningRateSchedule(self):
        def schedule(epoch_idx):
            return max(self.minlr, self.parent.learning_rate * pow(self.ReductionFactor, epoch_idx // self.ReduceEveryXEpoch))
            
        return LearningRateScheduler(schedule=schedule)
    
    def ReduceOnPlateau(self):
        # optional monitor val_loss if train/test split is activated
        return ReduceLROnPlateau(monitor='loss', factor = self.ReductionFactor, patience=self.ReduceEveryXEpoch, min_lr = self.minlr)
                                 
    def LearningRateCallBack(self):
        if self.Mode == lrMode.Constant:
            return None
        if self.Mode == lrMode.ReduceOnPlateau:
            return self.ReduceOnPlateau()
        if self.Mode == lrMode.Schedule:
            return self.LearningRateSchedule()